// Copyright 2009 Thomas Hoppe
// -----------------------------------------
// This work is licensed under the Creative Commons Attribution License. 
// To view a copy of this license, visit http://creativecommons.org/licenses/by/2.0/ 
// or send a letter to Creative Commons, 559 Nathan Abbott Way, Stanford, California 94305, USA.
// You are free:
// - to copy, distribute, display, and perform the work
// - to make derivative works
// - to make commercial use of the work
// Under the following conditions:
// - Attribution. You must give the original author credit.
// - For any reuse or distribution, you must make clear to others the license terms of this work.
// - Any of these conditions can be waived if you get permission from the copyright holder.
// Your fair use and other rights are in no way affected by the above. 

#declare GlobalIvySeed=seed(45634);

#declare LeafBody=
union               
{   
	#local Bri=rand(GlobalIvySeed)*.4+.1;	// brightness variance of leaf
	#local Rot0=rand(GlobalIvySeed)*40; // leaf angle variance
	#local RHandle=<0, .8, .5> + vrnd(0, 1.0, .4, GlobalIvySeed);

	// leaf body
	box 
	{ 
		<-.5, -1, -.005>, <.5, 0, .005> 

		texture
		{
			pigment { image_map { png "leaf2.png" map_type 0 interpolate 4 } translate x*.5 } 
			normal { bump_map { png "leaf2.png" bump_size .5 map_type 0 interpolate 4 } translate x*.5 } 
			finish { diffuse 1.0-Bri ambient Bri }
		}
		              
		translate y*.25
		rotate x*Rot0
		translate y*.25 
	}
     
   // leaf handle                                                  
	union
	{                                                           
		cylinder { <0, .25, 0>, <0, .6, .1>, .02 }
		cylinder { <0, .6, .1>, RHandle, .02 }     
		
		pigment               
		{ 
			marble	
			turbulence .2			
			color_map
			{
				[ 0.0  color rgb <.00, .25, .00> ]
				[ 0.9  color rgb <.00, .35, .00> ]
				[ 1.0  color rgb <.50, .65, .15> ]
			}   
			scale .01
		}			
	}
	
	// move, so <0, 0, 0> is the 'plug' pos, -Z should harmonize with Norm, +y with Start->End
	translate -RHandle
}


// Place ivy leaf at percent Pos betweeen Start and End, size Scale.
// Leaf orientation is calulated from Start/End and Norm.
#macro IvyLeafObj(IvySeed, Start, End, Norm, Pos, Scale)       
	// adjustable params:
	// -----------------------------------------------------------------------
	#local Rot=rand(IvySeed)*120-60;	// rotation variance twig axis
	#local Sz=rand(IvySeed)*.3+.7;	// leaf size variance  
	// -----------------------------------------------------------------------
	
	// transform to meet the desired location (Start/End/Norm/Pos)
	#local Vec=vnormalize(End-Start);
	#local Axis=vcross(Vec, y);
	#local Angle=degrees(acos(vdot(Vec, y)));
	 
	// reorient normal, so it is 90 to Start->End
	#local Norm=vnormalize(Norm);
	#local CrossSEN=vcross(Vec, Norm);               
	#local NNorm=-vcross(Vec, CrossSEN);
	// check if it points to the same direction as the original normal
	// otherwise fix it
	#if(vdot(Norm, NNorm)<0)
		#local NNorm=-NNorm;
	#end
	
	// orient leaf to normal
	#local Orient=vaxis_rotate(-z, Axis, -Angle);              
	
	// measure angle between orient and normal
	#local NAngle=degrees(acos(vdot(vnormalize(Orient), vnormalize(NNorm))));
	
	// find rotation direction by rotating orient in positive dir
	#local NOrient=vaxis_rotate(Orient, Vec, NAngle);
	// verify angle dir                                        
	#if(vdot(vnormalize(NNorm), vnormalize(NOrient))<.9)
		#local NAngle=-NAngle;
	#end                                             
	//#local NOrient=vaxis_rotate(Orient, Vec, NAngle);

	object
	{
		LeafBody
				
		//rotate vrnd(40, 40, 40, IvySeed)		
		
		// scale requested size
		scale Scale*Sz
		
		Axis_Rotate_Trans(Axis, -Angle)  
		Axis_Rotate_Trans(End-Start, NAngle+Rot)
		translate Start + Pos*(End-Start) 
	}  
#end                        

#declare DebugSeed=seed(34539);

#macro Ivy(Seed, GrowStart, StartDir, Obj, MaxLife, LenF, Thickness, Depth, ChanceToSplit, MinTwigDist, VarTwigDist, 
	RandomWeight, PreviousWeight, UpWeight, GravityWeight, LeafScale, LeafCount, GenerateLeaves, Progress, Debug)   
	// adjustable params
	#local SearchGrain=2;  // search resolution for clamping to obj.
	                       // value can increase parsing time heavily, 6*(n+1)^2 calls to trace !
	
	#local MinTwigLife=0.7; // percent of MaxLife guarenteed
	#local VarTwigLife=0.3; // percent of MaxLife variance
 	                
	union
	{     
		#local IvySeed=seed(Seed);  
		#local IvySplitSeed=seed(Seed+999);
		
		// lifetime of this twig
		#local Life=MaxLife*(MinTwigLife+VarTwigLife*rand(IvySeed));
		// first growing direction
		#local Last=GrowStart-StartDir;                    
		
		// search first intersection point
		#local Norm=<0, 0, 0>;
		#local Current=Clamp(GrowStart, Obj, SearchGrain, Norm, false);
		
		// must be valid to proceed
		#if(vlength(Norm)!=0)                  
		
			// place above intersection point               
			#local CurrentN=Current+Norm*(MinTwigDist+VarTwigDist*rand(IvySeed));

			// colorice every twig with randow color
			#if(Debug=true)
				#local DebugColor=<rand(DebugSeed), rand(DebugSeed), rand(DebugSeed)>;
			#end
			
			// grow this twig   
			#local N=0;
			#while(N<Life) 
			
				#if(Progress=true)
					#debug concat("Ivy-Progress: ", str(N, 0, 0), "/", str(Life, 0, 0), "\n")
				#else
					#debug "."
				#end       
				
				// calculate next growing pos, 4 directions:
				// random:
				#local Ran=false;
				#while(Ran=false)
					#local RDir=vnormalize(vrnd(1, 1, 1, IvySeed));
					// check if not opposite 
					#if(vdot(RDir, Current-Last)>0)
						#local Ran=true;
					#end
				#end
				// previous dir:
				#local PDir=vnormalize(Current-Last);
				// up dir:
				#local UDir=y;
				// and gravity:
				#local GDir=-y;
				          
				#local Dir=RandomWeight*RDir + PreviousWeight*PDir + UpWeight*UDir + GravityWeight*GDir;
				#local Dir=Dir*LenF;
				
				// Dir contains the position relative to our last position
				
				// find intersection to calculated pos
				#local Norm=<0, 0, 0>;  
				#local Deb=false;  
				/*				
				#if(N=29)
					#local Deb=true;
				#end       
				*/
			    #local Next=Clamp(CurrentN+Dir, Obj, SearchGrain, Norm, Deb);  
			    //#debug concat("N: ", str(N, 0, 0), " P: ", vstr(3, Current+Dir, ", ", 0, 2), " - I:", vstr(3, Next, ", ", 0, 2), " - Norm:", vstr(3, Norm, ", ", 0, 2), "\n")
			    #if(vlength(Norm)!=0)
			    
			    	// start + endpoint valid, proceed
			    	
			    	// place endpoint above intersection point
			    	#local NextN=Next+Norm*(MinTwigDist+VarTwigDist*rand(IvySeed));
			    	
			    	// draw twig   
			    	//sphere { CurrentN, 0.01 pigment { color Red } } 
			    	union
			    	{
			    		// reduce with progress
			    		#local _Thickness=Thickness*((1-N/Life)*.7+.3);
						cylinder { CurrentN, NextN, _Thickness }
						sphere { NextN, _Thickness }
						texture 
						{       
							#if(Debug=true)
								pigment
								{
									color rgb DebugColor
								}
								finish
								{
									ambient 1
									diffuse 0
								}
							#else
								pigment               
								{ 
									marble	
									turbulence .2			
									color_map
									{
										[ 0.0  color rgb <.47, .38, .28> ]
										[ 0.9  color rgb <.47, .38, .28> ]
										[ 1.0  color rgb <.70, .55, .45> ]
									}   
									scale .01
								}
							#end
						} 
					} 
	
					// generate leaves
					#if(GenerateLeaves=true)
						#local LeafSeed=seed(Seed+1000+N);				
						#local LOS=rand(LeafSeed)*LeafCount; // Leaves on stem.
						#local i=0;
						#while(i<LOS)
							IvyLeafObj(LeafSeed, CurrentN, NextN, Norm, rand(LeafSeed), LeafScale*((1-N/Life)*.3+.7))
							#local i=i+1;
						#end  
					#end
					
					#if(Debug=true)
 						sphere { NextN, 10 * Thickness * LeafScale*((1-N/Life)*.3+.8) pigment { color Green } }
					#end
					
					// process childs
					#if(Depth>0)       
						#if(ChanceToSplit>rand(IvySplitSeed))                        
							//sphere { CurrentN, 0.02 pigment { color Green } }
							#local _MaxLife=1.5*(Life-N); 
							#if(_MaxLife>3)
								Ivy(Seed + N, CurrentN, Dir, Obj, _MaxLife, LenF, _Thickness, Depth-1, ChanceToSplit, MinTwigDist, VarTwigDist, 
									RandomWeight, PreviousWeight, UpWeight, GravityWeight, LeafScale, LeafCount, GenerateLeaves, false, Debug)
							#end
						#end
					#end
				#else
					// finish if nothing is found (hit a bounding plane exactliy for instance)
					#debug "Dead end reached\n"
					#local N=Life;				
			    #end              
			    #local Last=Current;
				#local Current=Next;
				#local CurrentN=NextN;
				#local N=N+1;
			#end
		#end
	}
#end
